!
! cabmatrix --- ADC for solids based on a Wannier Hartree-Fock calculation
! Christian Buth
! 10.02.2004
!
!
! Copyright (c) 2004-2006 Christian Buth
!
! This program is free software; you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation; either version 2 of the License, or
! (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
!
!
! Version 1.00: 10.02.2004 (Took Greens code from Martin)
! Version 1.10: 13.02.2004 (Quick and dirty start version with full diagonalization
!       employed for the band structure matrix)
! Version 1.20: 28.02.2004 (Active orbitals can be specified for correlation calculations)
! Version 1.30: 20.06.2004 (Separated diagonalization code, matrix written to disk)
!


!
! Program
!

      program cabmatrix
      use physconst
      use wannier
      use adcbands
      implicit none

      character(1024) fileFockip,     &                 ! Negative of Hartree-Fock orbital energies
     &                fileFockBands,  &                 ! FOCKALL Hartree-Fock band structure
     &                fileFockAll,    &                 ! Wannier direct space Fock matrices
     &                fileFockTrans,  &                 ! Wannier Fock matrix of the integral transformed unit cells
     &                fileVBucket,    &                 ! Wannier two-electron bucket file
     &                fileVijkl,      &                 ! Wannier two-electron integrals
     &                fileVIndex                        ! Wannier two-electron index file

      integer numUnitCells, startOccupied, endVirtual, periodicity, numBrillouinPaths
      integer, parameter :: ffockip    = 50, &
     &                      fadcip     = 51, &
     &                      ffockbands = 52, &
     &                      fadcbands  = 53

      type(twannier) wannbands
      type(tcoadc) coadcmat

      namelist /coadcmatrix/ numUnitCells, startOccupied, endVirtual, periodicity, &
     &          numBrillouinPaths, fileFockip, fileFockBands,        &
     &          fileFockAll, fileFockTrans, fileVBucket, fileVijkl,  &
     &          fileVIndex


      ! Greeting
      write (*,*) 'Crystal orbital algebraic diagrammatic constuction'
      write (*,*) 'Copyright (c) 2004-2006 Christian Buth'
      write (*,*)
      write (*,*)

      numUnitCells  = 1
      startOccupied =  1
      endVirtual    = -1
      periodicity   = 0

      fileFockip     = 'fockip.dat'
      fileFockBands  = 'fockbands.dat'
      fileFockAll    = 'FOCKALL0i1.DAT'
      fileFockTrans  = 'FTRANSA0i1.DAT'
      fileVBucket    = 'VBFINT0i1.BKT'
      fileVijkl      = 'VijklD0i1.DAT'
      fileVIndex     = 'VINDEX0i1.DAT'

      read (*,coadcmatrix)

      ! Initialize CO-ADC
      call initwannier(wannbands, periodicity, numUnitCells, &
     &          fileFockTrans, fileVBucket, fileVijkl, fileVIndex, &
     &          fileFockAll)

      if (endVirtual < 0) then
        endVirtual = wannbands%numOrbitals
      end if

      call initcoadc(coadcmat, wannbands%numOrbitals, wannbands%numOccupied, &
     &               startOccupied, endVirtual, wannbands%numUnitCells)

      ! Carry out CO-ADC calculation
      select case (wannbands%periodicity)
        case (0)
          write (*,*) 'The system is a molecule'
        case (1)
          write (*,*) 'The system is a polymer'
        case (2)
          write (*,*) 'The system is a surface'
        case (3)
          write (*,*) 'The system is a crystal'
      end select

      if (wannbands%periodicity == 0) then
        call ipea
      else
!!!        call twointdecay
        call bandstruct
      end if

      ! Clean up CO-ADC
      call exitcoadc(coadcmat)
      call exitwannier(wannbands)

      write (*,*)
      write (*,*)
      write (*,*)
      write (*,*) 'Excecution sucessful'
      write (*,*)
      write (*,*) 'Da steh ich nun, ich armer Tor!'
      write (*,*) 'Und bin so klug als wie zuvor;'
      write (*,*) '                Goethe, Faust I'


      contains

!
! ipea --- calculate the Hartree-Fock and ADC ionization potentials and electron affinities
!

      subroutine ipea
      implicit none

      logical exactoffdiag, fockpt3

      integer i

      real(dblprec), allocatable :: fockeigenvectors(:,:), &             ! Eigenvectors of the Fock matrix
     &                              fockeigenvalues(:)                   ! Eigenvalues of the Fock matrix

      ! The Fock matrix of the molecule
      allocate(fockeigenvectors(wannbands%numOrbitals,wannbands%numOrbitals), &
     &         fockeigenvalues(wannbands%numOrbitals))
      write (*,*) 'Calculating the Hartree-Fock ionization potentials and electron affinities...'
      open(unit=ffockip,file=fileFockip,status='replace',action='write')
      call molfock(wannbands, 'V', fockeigenvectors,fockeigenvalues)
      do i=startOccupied, endVirtual
        write (ffockip,'(2F15.6)') -fockeigenvalues(i) * HartreetoeV, 1.0
      end do
      write (ffockip,*)
      close(unit=ffockip)

      ! ADC ionization potentials and electron affinities
      write (*,*) 'Calculating the ADC matrix...'
      exactoffdiag = .true.
      exactoffdiag = .false.
      fockpt3 = .false.
      call adcmatrix(wannbands, coadcmat, fockeigenvectors, fockeigenvalues, exactoffdiag, fockpt3)
      deallocate(fockeigenvectors,fockeigenvalues)
      end subroutine ipea


!
! bandstruct --- calculate the Hartree-Fock and CO-ADC band structures
!

      subroutine bandstruct
      use bandstructure
      implicit none

      logical exactoffdiag, fockpt3
      integer currPath, currkpoint
      real(dblprec), allocatable :: eigenvalues(:)                      ! Eigenvalues of the Fock matrix
      real(dblprec), allocatable :: dummy(:,:)
      type(tlatticevector) kpoint
      type(tfbands) fbands

      ! Get paths though the Brillouin zone
      allocate(fbands%paths(numBrillouinPaths))
      fbands%fock = .true.
      fbands%numPaths = numBrillouinPaths
      fbands%startOccupied = 1
      fbands%numOccupied = wannbands%numOccupied
      fbands%numOrbitals = wannbands%numOrbitals
      fbands%periodicity = wannbands%periodicity
      fbands%spacegroup  = 0
      do currPath=1, numBrillouinPaths
        call readbrillouinpathspec(fbands%paths(currPath), fstdin, wannbands%periodicity, 0)
      end do
      call reportbrillouinpathspec(fbands%paths, numBrillouinPaths, wannbands%periodicity)

      ! Hartree-Fock band structure
      allocate(eigenvalues(wannbands%numOrbitals))

      write (*,*) 'Diagonalizing the Fock matrix...'
      call openbands(fbands, ffockbands, fileFockBands, 'write')

      do currPath=1, numBrillouinPaths
        call putbrillouinpath(fbands, currPath)
        kpoint = fbands%paths(currPath)%beg
        do currkpoint=1, fbands%paths(currPath)%numkpoints
          call wannierfock(wannbands, kpoint, eigenvalues)
          eigenvalues = eigenvalues * HartreetoeV
          call putbandrecord(fbands, kpoint, wannbands%numOrbitals, eigenvalues)
!          kpoint%a = kpoint%a + fbands%paths(currPath)%inc%a
!          kpoint%b = kpoint%b + fbands%paths(currPath)%inc%b
!          kpoint%c = kpoint%c + fbands%paths(currPath)%inc%c
          kpoint%v = kpoint%v + fbands%paths(currPath)%inc%v
        end do
      end do

      call closebands(fbands)
      write (*,*) ' Hartree-Fock bands are in ',trim(fileFockBands)
      write (*,*)

      ! Correlated band structure matrix
      write (*,*) 'Calculating the CO-ADC band structure matrix...'
      exactoffdiag = .false.
      fockpt3 = .false.
      call adcmatrix(wannbands, coadcmat, dummy, eigenvalues, exactoffdiag, fockpt3)
      deallocate(eigenvalues)
      end subroutine bandstruct
      end program cabmatrix
