!
! zbnddia --- diagonalizes the complex symmetric band matrix
!       computed by the complex symmetric band Lanczos
! Thomas Sommerfeld based on a real symmetric code by Hans-Dieter Meyer
! Planet PKS version by Christian Buth
! 26 July 2004
!
!
! Copyright (c) 2003 Hans-Dieter Meyer
! Copyright (c) 2003 Thomas Sommerfeld
! Copyright (c) 2004-2006 Christian Buth
!
! This library is free software; you can redistribute it and/or
! modify it under the terms of the GNU Lesser General Public
! License as published by the Free Software Foundation; either
! version 2.1 of the License, or (at your option) any later version.
!
! This library is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! Lesser General Public License for more details.
!
! You should have received a copy of the GNU Lesser General Public
! License along with this library; if not, write to the Free Software
! Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
!
!
!
!     diagonalizes the bandmatrix computed by the block lanczos
!     and purge the spectrum
!
!     this is not difficult but the purging and output selection
!     is quite complex; in addition there are many fossil structures
!     of Dieter or Uwe and some of the aux arrays are used for several
!     purposes
!
!     n        matrix dimension = no. of lanczos iterations
!     np       half band width  = blocksize in lanczos
!     t        bandmatrix
!     z        contains the first and last np components
!              of the eigenvectors of t
!     a,e,e2   aux arrays
!     d        contains the eigenvalues of t
!     ps       contains the polestrengths (Sommerfeld-Meyer definition)
!     thresh1  threshold for eigenvalue output
!     thresh2  threshold for output of coefficients
!     thresh1 > thresh2 is assumed
!
!     the following parameters are needed to compute eigenvectors.
!     in the 1st version all eigenvectors of t were computed for
!     this purpose and z(n,n); in this version the eigenvectors are
!     computed using inverse iteration and z(np2,n), i.e., zdim is
!     always np2 = 2 * np
!
!     neigenv          no. of eigenvectors
!     evlist           is the list of eigenvectors to be computed;
!                      if evlist(1)=0 the eigenvectors are selected
!                      via a max overlap criterion (default)
!     au, al, b, ipvt  are aux arrays for the inverse iteration
!                      they are not referred if neigenv = 0
!     iparam is an array with flags
!     iparam(1) = dieterout (output in dieters format)
!
!     the old output of dieter (first and last n eigenvalues) is
!     controlled via dieterout; my output (all eigenvalues with a
!     certain overlap with the startspace) is controlled via
!     thresh1
!     -------------------------------------------------------------
!
      subroutine zbnddia(n, np, t, d, e2, z, thresh, neigenv,  &
     &          evlist, iparam, jnd, ps)
      use lanczos
      implicit none

      integer, intent(in) :: n, np, iparam(2)
      integer, intent(out) :: jnd
      integer, intent(inout) :: neigenv, evlist(neigenv)
      real(dblprec), intent(in) :: thresh(3)
      real(dblprec), optional, intent(out) :: ps(n)
      complex(dblprec), intent(in) :: t(np+1,n)
      complex(dblprec), intent(out) :: d(n), e2(n), z(2*np, n)

! CB invit
      integer, allocatable :: ipvt(:)
      complex(dblprec), allocatable :: au(:,:), al(:,:), b(:)
! END CB invit
      complex(dblprec), allocatable :: a(:,:), e(:)

      integer maxn, maxnp
      parameter (maxn = 30000, maxnp = 500)

      integer np1, np2, zdim, i, j, iwgt(maxnp), nwgt
! CB invit
      integer ninvi, mmm
! END CB invit
      integer ind2(maxn), indx, ngt, wrterf
      real(dblprec) wgt(maxnp)
      real(dblprec) thresh1, thresh2, eta

!
!     Dieters Programmteile (war vorher implizit deklariert)
!
      integer dieterout
      integer nsp, nmu, isp, ind(maxn), mlt(maxn), lsp(maxn)
      integer k, kk, ia, ic, jc, icnv, jcnv
      complex(dblprec) aa, sum, sup, ak, a1k, xm
      real(dblprec) ovl, thrs, y1, y2, x1, x2

      complex(dblprec) zdotu
      external zdotu, zbandr, ztql2

! CB invit
      if (neigenv /= 0) then
        allocate(au(n,2*np+1), al(n,2*np+1), b(n), ipvt(n))
      end if
! END CB invit
      allocate(a(n,np+1), e(n))


      np1 = np + 1
      np2 = 2 * np
      zdim = np2
! CB invit
      mmm = np2 + 1
      ninvi = 1
! END CB invit
      dieterout = iparam(1)
      wrterf    = iparam(2)
      thresh1   = thresh(1)
      thresh2   = thresh(2)
      eta       = thresh(3)

      write (6,4000)
 4000 format (//'  Diagonalization of the band-matrix T (complex ',  &
     &          'symmetric)'/)
      write (6,*) '    matrix dimension : ', n
      write (6,*) '    half band width  : ', np
      if (n .gt. maxn .or. np .gt. maxnp) then
         write(6,*)
         write(6,*) 'maxdim = ', maxn
         write(6,*) 'maxnp  = ', maxnp
         write(6,*) 'change parameters in zbnddia.'
         stop 'zbnddia: change parameters in zbnddia!'
      end if
!     -----------------------------------------------------
!     t in bandrc format store in new format (transposed)
!     z could be used for this purpose, but t is considered
!     to be small in comparison to the lanczos blocks
!     -----------------------------------------------------
      do i = 1, n
         do j = 1, np1
            a(i,j) = t(j,i)
         end do
      end do
!
!     zbandr is called anyway to initialize z
!
      call zbandr(zdim, n, np, a, .true., z)
!
!     Rearrange the tridiagonal matrix
!
      d(1) = a(1,1)
      e(1) = (0.0D0, 0.0D0)
      do i = 2, n
         d(i) = a(i,1)
         e(i) = a(i-1,2)
      end do

      call ztql2(zdim, n, d, e, z, i)
      if (i .ne. 0) then
         write (6,*) '===  Error in ztql2  ==='
         write (6,*) i,  &
     &        '-th eigenvalue not converged after 30 iterations'
         stop 'zbnddia: error in ztql2'
      end if
!
!     write the first unpurged eigenvalues to stdout
!
      write (6,'(/,A)') '  The lowest 10 eigenvalues of T:'
      do i = 1, min(10, n)
         write (6,3000) i, d(i)
 3000    format (4x,I3,'  (',F15.8,',',F15.8,')')
      end do
!
!     T has been diagonalized, now purge the spectrum and
!     prepare a nice output
!
!     ----------------------------------------------------------
!     DETERMINE THE INDEX-VECTOR 'IND' IN ORDER TO REMOVE
!     MULTIPLE AND SPURIOUS EIGENVALUES.
!     ----------------------------------------------------------
      NSP = 0
      NMU = 0
      ISP = 0
      JND = 0
      J   = 0

  50  J   = J + 1
      IF( J .GT. n )                               GOTO  40
!      CALL zMULT( Z(1,J), Z(1,J), NP , AA )
      aa=zdotu(np,z(1,j),1,z(1,j),1)
      NSP = NSP + 1
      ISP = ISP + 1
!.....TEST OF SPURIOUS EIGENVALUE.
      IF( zabs(AA) .LT. 1.D-14 )                     GOTO  50
      NSP = NSP - 1
      ISP = ISP - 1
      JND = JND + 1
      IND(JND) = J
      MLT(JND) = 0
      LSP(JND) = ISP
      ISP      = 0

      IF( J .EQ. n )                               GOTO  40
      THRS = 5.D-7*( 1.D0 + zabs(d(J)) )
!.....TEST OF MULTIPLE EIGENVALUE.
      IF( zabs( d(J) - d(J+1) ) .GT. THRS )           GOTO  50
      SUM = AA
      SUP = AA
      KK  = 0
      IA  = 0

      do k = 1, n - j
      OVL = 1.D-5
      IF( zabs( d(J+IA) - d(J+K) ) .GT. THRS )        GOTO  42
!      CALL zMULT( Z(1,J+K), Z(1,J+K), NP , AK )
      ak = zdotu(np,z(1,j+k),1,z(1,j+k),1)
      IF( zabs(AK) .LE. 1.D-14 )  THEN
         KK     = K
         cycle
      ELSE IF( zabs(AK/SUP + SUP/AK)  .LT.  2.D+6 )    THEN
           OVL = 1.D-7
           IF( zabs(d(J+IA)-d(J+K)) .GT. 1.D-2*THRS )   GOTO  42
      END IF
!      CALL zMULT( Z(1,J+K), Z(1,J+IA), NP , A1K )
      a1k=zdotu(np,z(1,j+k),1,z(1,j+ia),1)
      A1K = A1K*A1K/(SUP*AK)
      IF( zabs( A1K-1.D0 ) .LE. OVL )            THEN
         SUM = SUM + AK
         IF( zabs(AK) .GT. zabs(SUP) )  THEN
           SUP = AK
           IA  =  K
         END IF
         KK = K
      ELSE
        GOTO  42
      END IF
      end do

  42  continue
      d(J) = d(J+IA)
      do i = 1, np
         Z(   I,J) = Z(   I,J+IA)*zSQRT(SUM/SUP)
         Z(NP+I,J) = Z(NP+I,J+IA)
      end do

      NMU = NMU + KK
      J   =  J  + KK
      MLT(JND) =  KK
      GOTO  50

  40  CONTINUE
!
!.....Remove spurious and multiple eigenvalues
!
      do j = 1, jnd
         I = IND(J)
         d(J) = d(I)
         do k = 1, np2
            z(k,j) = z(k,i)
         end do
      end do

      write (6,'(/,A)') '  Information on purging:'
      PRINT 905, JND, n, int(100.0D0*dble(jnd)/dble(n))
 905  FORMAT(4x,'There are',I5,' relevant residues out of',I5,  &
     &     '   (',I3,'%)')
      write(6,'(4x,A,I4,6x,A,I4)') 'N_spurios = ', NSP,  &
     &     'N_multiple = ', NMU

!
!.....TRANSFORM THE EIGENVECTORS(FIRST NP ROWS) TO OBTAIN THE RESIDUES.
!.....STORE THE RESULT IN THE FIRST NP ROWS OF THE EIGENVECTOR MATRIX Z.
!.....MULTIPLY :  EIGENVECTORS(LAST ROWS) * 'ERROR-TRIANGLE'
!.....STORE THIS RESULT IN THE LAST ROWS OF THE EIGENVECTOR MATRIX Z.
!
      do j  = 1, jnd
        do i  = 1, np
          XM = (0.0D0, 0.0D0)
          do k = i, np
            KK = K + NP
            IC = K + n - NP
            JC = I - K + NP1
            XM = XM + Z(KK,J) * T(JC,IC)
          end do
          Z(I+NP,J) = XM
        end do
      end do
!
!.....PRINT MEAN RESIDUES (X), EIGENVALUES (W), ERROR-EST.(Y) AND MULT.
!
      if (dieterout .eq. 0) GOTO 329
      ICNV = 0
      JCNV = 0
      PRINT 900
 900  FORMAT(//2X,'E I G E N V A L U E S'/2X,125('-')//2  &
     &     ('   NO.  MEAN-RES.*100  EIGEN-VALUE',16x      &
     &     ,'ERROR     M   S',3x) // 2X,125('-')/ )
      do j = 1, (jnd+1)/2
         Y1 = 0.D0
         Y2 = 0.D0
         K  = JND + 1 - J
         do i = 1, np
            Y1 = Y1 + dble(Z(NP+I,J) * dconjg(z(np+i,j)))
            Y2 = Y2 + dble(Z(NP+I,K) * dconjg(z(np+i,k)))
         end do
         Y1 = dsqrt(Y1)
         Y2 = dsqrt(Y2)
         IF(Y1.LT.1.D-9 .AND. J.NE.JND/2+1) ICNV = ICNV + 1
         IF(Y2.LT.1.D-9)  ICNV = ICNV + 1
         IF(Y1.LT.1.D-6 .AND. J.NE.JND/2+1) JCNV = JCNV + 1
         IF(Y2.LT.1.D-6)  JCNV = JCNV + 1
         IF( J .GT. dieterout ) cycle
         X1 = 0.D0
         X2 = 0.D0
         do i = 1, np
            X1 = X1 + dble(Z(I,J) * dconjg(z(i,j)))
            X2 = X2 + dble(Z(I,K) * dconjg(z(i,k)))
         end do
         X1 = DSQRT(X1)
         X2 = DSQRT(X2)
         PRINT 901, J,X1,d(J),Y1,MLT(J),LSP(J), K,X2,d(K),Y2,MLT(K)  &
     &        ,LSP(K)
 901     FORMAT (2(I5,F8.4,' (',F14.8,',',E15.7,')',E9.2,I3,1X,I3))
      end do
      PRINT 990
 990  FORMAT(/2('   NO.  MEAN-RES.      EIGEN-VALUE',16x,  &
     &'ERROR     M   S', 4X) // 2X,125('-')/ )
      PRINT 902, ICNV, 100.D0*ICNV/DBLE(JND), 100.D0*ICNV/DBLE(n)
      PRINT 903, JCNV, 100.D0*JCNV/DBLE(JND), 100.D0*JCNV/DBLE(n)
 902  FORMAT(/,'    NUMBER OF CONVERGED (1.D-9) EIGENVALUES =',  &
     &         I4, F11.2, ' %  (JND) ', F9.2,' %  (JEND) ' )
 903  FORMAT(/,'    NUMBER OF CONVERGED (1.D-6) EIGENVALUES =',  &
     &         I4, F11.2, ' %  (JND) ', F9.2,' %  (JEND) ' )
 329  CONTINUE
!     --------------------------------------------------
!     print eigenvalues mit mean-res > thresh1
!     collect the corresponding indices in ind2()
!     and the mean res in the array e2  (aux for zbandr)
!     --------------------------------------------------
      write(6,400) thresh1
 400  format(//2x,68('-')/'  Eigenvalues with mean-res > ',F6.3,/      &
     &  /'   no.  mean-res.       eigenvalue',16x,'error estimate'/2x  &
     &     ,68('-'))
      ngt = 0
      do j = 1, jnd
         x1 = 0.0D0
         do i = 1, np
            x1 = x1 + dble(z(i,j) * dconjg(z(i,j)))
         end do
         x1 = dsqrt(x1)
! CB polestrengths zurckgeben (Sommerfeld-Meyer definition)
         if (present(ps)) then
           ps(j) = x1
         end if
! END CB polestrengths
         if (x1 .gt. thresh1) then
            ngt = ngt + 1
            ind2(ngt) = ind(j)
            e2(ngt) = dcmplx(x1, 0.0D0)
            y1 = 0.0D0
            do i = 1, np
               y1 = y1 + dble(z(np+i,j) * dconjg(z(np+i,j)))
            end do
            y1 = dsqrt(y1)
            write(6,420) j, x1, d(j), y1, ind(j)
 420        format(I5,2x,F8.4,'  (',F14.8,',',1PE16.8,')',1PE11.2,3x,'('  &
     &           ,I4,')')
         end if
      end do
!     ---------------------------------------------------------
!     write the eigenvalue list to a separate file for eta-runs
!     ---------------------------------------------------------
      if (wrterf .eq. 1) then
         open (ifile,file='ETARUN',access='SEQUENTIAL',form='FORMATTED')
         write(ifile,'(1PE18.8)') eta
         write(ifile,'(I6)') ngt
         do j = 1, ngt
            write(ifile,'(I4,1PE18.8,1PE18.8)') j, d(ind2(j))
         end do
         close (ifile)
      end if
!     ---------------------------------------------------------
!     compute eigenvectors of Z using inverse iteration
!     the eigenvectors are written to disk: ZMAT
!     if evlist is undefined the first neigenv in ind2 are used
!     ---------------------------------------------------------
      if (neigenv .gt. 0) then
         if (evlist(1) .eq. 0) then
            if (neigenv .gt. ngt) then
               write(6,*) 'Only ', ngt, ' vectors can be computed.'
               neigenv = ngt
            end if
            do i = 1, neigenv
               evlist(i) = ind2(i)
            end do
         end if
         write(6,430) (evlist(i), i = 1, neigenv)
 430     format(/,'  The following eigenvectors are computed:'/,  &
     &        20(4x,10(2x,I5),/))
         open (ifile,file='ZMAT',access='sequential',form='unformatted')
         write(ifile) n, neigenv
         do j = 1, neigenv
            do i = 1, n
               b(i) = (0.0D0, 0.0D0)
            end do
            do i = 1, np
               b(i)          = z(i, evlist(j))
               b(n - np + i) = z(np + i, evlist(j))
            end do
            indx = 0
            do i = 1, jnd
               if (evlist(j) .eq. ind(i)) then
                  indx=i
               end if
            end do
            if (indx .eq. 0) then
               write(6,*) 'Index ', evlist(j), ' has not been found.'
               stop 'zbnddia: index has not been found!'
            end if
! CB invit
            call zinvit(n,np,np1,mmm,d(indx),t,au,al,b,ipvt,  &
     &           ninvi,.false.)
!            stop 'no invit!'
! END CB invit
            write(ifile) (b(i), i = 1, n)
         end do
         close(ifile)
      end if
!     -----------------------------------------------------
!     print weights of start vectors for mean-res > thresh2
!     -----------------------------------------------------
      if (thresh2 .lt. 1.0d0) then
         write(6,450) thresh2
 450     format(/2x,60('-')/'  absolute weights (|C|**2) for m-r >',F6.3  &
     &        ,/)
         do i = 1, jnd
            x1 = 0.D0
            do j = 1, np
               x1 = x1 + dble(z(j,i) * dconjg(z(j,i)))
            end do
            x1 = dsqrt(x1)
            if (x1 .gt. thresh2) then
               nwgt = 0
               do j = 1, np
                  x1 = dble(z(j,i) * dconjg(z(j,i)))
                  if (x1 .gt. 0.001d0) then
                     nwgt = nwgt + 1
                     iwgt(nwgt) = j
                     wgt(nwgt) = x1
                  end if
               end do
               write(6,470) i, d(i)
 470           format(1X,I4,'-th root  E = (',F14.8,',',F14.8,')')
               write(6,480) (wgt(j), iwgt(j), j=1, nwgt)
 480           format(6(4x,8(2x,F5.3,'(',I2,')'),/))
               write(6,*)
            end if
         end do
      end if
!
!.. ...PRINT MEAN THE 'ERROR TRIANGLE' OF THE T-MATRIX.
!
!      PRINT 940
! 940  FORMAT(//2X,'" E R R O R - T R I A N G L E "  (TRANSPOSED)'/
!     &       /2X,125('-')//)
!      DO  333  II = 1, NP
!         PRINT 941, (T(NP1+K-II,n+1-K),K=1,II)
! 941  FORMAT(1X, 10F12.6 )
!         IF( II .GE. 10 ) PRINT *
! 333  CONTINUE
!

! CB invit
      if (neigenv /= 0) then
        deallocate(au, al, b, ipvt)
      end if
! END CB invit
      deallocate(a, e)
      end subroutine zbnddia
