!
! hblanc --- band Lanczos for complex Hermitian matrices
! Hans-Dieter Meyer
!
!
! Copyright (c) 2003 Hans-Dieter Meyer
! Copyright (c) 2003 Thomas Sommerfeld
! Copyright (c) 2004-2006 Christian Buth
!
! This library is free software; you can redistribute it and/or
! modify it under the terms of the GNU Lesser General Public
! License as published by the Free Software Foundation; either
! version 2.1 of the License, or (at your option) any later version.
!
! This library is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! Lesser General Public License for more details.
!
! You should have received a copy of the GNU Lesser General Public
! License along with this library; if not, write to the Free Software
! Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
!
!
c     ----------------------------------------------------------------
c
c     BAND LANCZOS for complex Hermitian matrices
c
c     large parts of this routine are taken from the block lanczos
c     program by H-D. Meyer and U. V. Riss
c
c     a matrix A is transformed to band-diagonal form T
c     A is only refered to via the subroutine hmatblk, which
c     has to perform the matrix times vector-block operation
c
c     ndim     dimension of the matrix A
c     np       block-size for the iteration --> band width of T
c     niter    no. of iterations (matrix*vector operations)
c     neigenv  No of eigenvectors to be computer, zero for no
c              second eigenvector run
c     ncartsv  If > 0, Lanczos vectors are orthogonalized to the
c              cartesian starting vectors, i.e. it is assured
c              that the element of the Lanczos vectors vanishes
c              where a cartesian start vector has its only non-zero
c              entry which is unity
c     iparam   array of additional parameters
c     iparam(1)  jstart     1st iteration  (can be > 1 in restarts)
!! CB REMOVED c     iparam(2)  jend       index of last iteration (output)
c     iparam(3)  neigenv    abs of no. of eigenvectors to be computed
c                neigenv =  0  : normaler Lanczos
c                neigenv = -1  : Vorlauf fr Eigenvektoren
c                neigenv >  0  : Eigenvektorlauf
c     iparam(4)  restart
c     iparam(5)  wrtrsf    write a restart file
c     iparam(6)  wrttmat   write T to a file
c     iparam(7)  ncartsv   no. of cartesian start-vectors
c                if the start-vectors are cartesian this is used
c                to normalize the lanczos-vectors on the start-block
c
c     t        contains on output the computed band matrix
c     p        contains 3*np vectors needed in the iteration
c              on input p(1,i,2) is used as the i-th start-vector
c              (it's not my fault)
c
c     z and ev are only referenced, if iparam(3) > 0
c     z        contains some eigenvectors of T from a previous run
c     ev       contains on output the corresponding eigenvectors of A
c
c     icartsv  contains the indices of the cartesian start-vectors
c
c     external hmatblk (ndim, np, pold, pnew, work, jdiesel)
c     is the mat times vector routine
c     work contains space for zmatvec and is not referenced in the lanczos
c
c     THE SUBROUTINE INDEXL COMPUTES THE CURREND INDEX OF THE LANCZOS-
c                           VECTOR CONTAINED IN THE ARRAY P.
c
c     -----------------------------------------------------------------
c
      subroutine hblanc(lan, status)
!      subroutine hblanc(ndim, np, niter, neigenv, ncartsv, iparam, t,
!     &     p, z, ev, icartsv, am) !!, hmatblk, mtxwork, jdiesel)
      use lanczos
      implicit none

!      integer ndim, np, niter, neigenv, ncartsv
!      integer iparam(10), icartsv(ncartsv)
!      complex*16 p(ndim,np,3), t(np+1,niter)
!      complex*16 z(niter,neigenv), ev(ndim,neigenv)
!      type(tcoadcmatrix), intent(inout) :: am
      integer, intent(out) :: status
      type(tlanczos), intent(inout) :: lan

!      integer jstart
!      integer restart, wrtrsf, wrttmat

      real(dblprec) r
      complex(dblprec) y
      integer i, j, k   !, nrun, ndone
      integer ii, jn, in, jo, io, jjc, iic, im, ip, jp
!      integer jiter, jko
!      logical da

      complex(dblprec) zdotc
      external zaxpy, zdotc, zdscal

      include 'iosetting.inc'

c     -------------------------------------------------------------
c     preliminaries
c     -------------------------------------------------------------
!      write (6,100)
! 100  format (//,2X,50('-')/,2X,'Band-Lanczos (complex '
!     &        'Hermitian)'/)
!      jstart  = iparam(1)
!      restart = iparam(4)
!      wrtrsf  = iparam(5)
!      wrttmat = iparam(6)
!      jstart = 1

c
c     Main Loop of the Lanczos Iteration
c     ----------------------------------
!      nrun = niter - jstart + 1
!      ndone = 10

      ! Do the respective number of band Lanczos iterations
      !do JJ = JSTART, NITER
      if (lan%jj > NITER) then
        status = planczosfinished
        return
      end if

         ! in -- index of block Lanczos vector of the next iteration
         ! jn -- remaining Lanczos vectors of the block to be processed
         ! cyclic(in+1) -- index of the current block Lanczos vector
         call indexl(lan%jj - lan%jstart + lan%np, lan%np, lan%jn, lan%in)
c
c     when neccessary do the next matrix times block operation
c
         if (mod(lan%jj - lan%jstart, lan%np) == 0) then
            ! io -- index of the result of the block matrix times matrix operation
            lan%io = mod(lan%in + 1, 3) + 1
        !    write (*,*) 'Block matrix times vector product No. ', JJ
        !    write (*,*) 'in: ', in, ' io: ', io
!           call hmatblk(ndim,np,p(1,1,io),p(1,1,in),mtxwork,jdiesel)
!            call bandmatvec(am, np, p(1:ndim, 1:np, io),
!     &                              p(1:ndim, 1:np, in))
           status = pmatblk
           return
         end if
c
c     Band-Lanczos algorithm as implemented by H-D. Meyer
c
         ! The lower triangle of the band matrix is computed and stored
         ! in packed storage format
         ! Compute intermediate vectors r in Eq. (3.5a)
         do ii = max(lan%jj - lan%np, 1), lan%jj - 1
            ! io -- previous block Lanczos vector
            ! jo -- previous Lanczos vector in the block
            call indexl(ii - lan%jstart + 3*lan%np, lan%np, jo, io)
            ! T(JJ, II) in packed storage (lower triangle)
            jjc = jj - ii + 1
c            do K = 1, NDIM
c               P(k,JN,IN) = P(k,JN,IN) - P(k,JO,IO)*T(JJC,II)
c            end do
            call zaxpy(lan%ndim, -lan%Tmat(jjc, ii), lan%lanvecs(1, jo, &
     &                 io), 1, lan%lanvecs(1, lan%jn, in), 1)
         end do

         do ii = lan%jj, lan%jj + lan%np - 1
            call indexl(ii - lan%jstart, lan%np, jo, io)
c            Y = (0.0d0, 0.0d0)
c            do K = 1, NDIM
c               Y = Y + conjg(P(k,JO,IO)) * P(k,JN,IN)
c            enddo
            ! Compute banddiagonal matrix elements in Eq. (3.5b)
            y = zdotc(ndim, lan%lanvecs(1, jo, io), 1, &
     &                      lan%lanvecs(1, lan%jn, lan%in), 1)
            ! T(II, JJ) = Y in packed storage (lower triangle)
            iic = ii - lan%jj + 1
            ! CB T(IIC, JJ) = Y
            T(iic, lan%jj) = conjg(y)
c            do K = 1, NDIM
c               P(k,JN,IN) = P(k,JN,IN) - P(k,JO,IO)*Y
c            end do
            ! Compute new vector r in Eq. (3.5c)
            call zaxpy(lan%ndim, -y, lan%lanvecs(1, jo, io), 1, &
     &                               lan%lanvecs(1, lan%jn, lan%in), 1)
         enddo
c     -------------------------------------------------
c     orthogonalization with respect to cartesian start-vectors
c     Vorsicht bei restart; icartsv muss definiert sein
c     -------------------------------------------------
         if (ncartsv .gt. 0) then
            do k = 1, np
               p(icartsv(k), lan%jn, lan%in) = 0
            enddo
         endif
c     --------------------
c     normalize new vector
c     --------------------
c         r = 0.0d0
c         do K = 1, NDIM
c            r = r + |P(k,JN,IN)|**2
c         enddo
         ! Compute last matrix element of the banddiagonal matrix in Eq. (3.5d)
         r = real(zdotc(lan%ndim, lan%lanvecs(1, lan%jn, lan%in), 1, &
     &                            lan%lanvecs(1, lan%jn, lan%in), 1), dblprec)
         r = sqrt(r)
         ! Packed storage lower triangle
         ! T(JJ, JJ + ndim) = Tpacked(JJ + ndim - JJ + 1, JJ)
         T(lan%np + 1, lan%jj) = conjg(cmplx(r, 0.0D0, dblprec))
         r = 1.0d0 / r
c         do K = 1, NDIM
c            P(k,JN,IN) = P(k,JN,IN) * r
c         end do
         ! Scale vector r in Eq. (3.5e)
         ! Block Lanczos vector for the next iteration is in block lan%in
         ! vector number lan%jn
         call zdscal(lan%ndim, r, lan%lanvecs(1, lan%jn, lan%in), 1)
c     -------------------------------------------------------
c     sum contributions of this iteration to the eigenvectors
c     -------------------------------------------------------
!         if (neigenv .gt. 0) then
!            jiter = np + jj
c
c     in the eigenvector calculation the very last np iterations
c     do not contribute to the eigenvectors;
c     but in case of a later restart these iterations have to be done
c
!            if (jiter .le. niter) then
!               do i = 1, neigenv
!                  do jko = 1, ndim
!                     ev(jko,i)=ev(jko,i)+z(jiter,i)*p(jko,jn,in)
!                  enddo
!               enddo
!            endif
!         endif
c
c     check how much has been done
c
!         if ((jj-jstart+1)*100 / nrun .ge. ndone) then
!            write(6,'(6X,I3,A)') ndone, '%'
!            ndone = ndone + 10
!         endif

!      end do
      lan%jj = lan%jj + 1
      status = piterdone
c     ------------------------------------------------------------
c     End of the Block-Lanczos iteration
c     ------------------------------------------------------------
!      write(6,999)
! 999  FORMAT(/,2X,'Lanczos finished',/,2X,50('-'))


      contains

c
c     one of Dieter's more serious crimes:
!     jj -- (in) current band Lanczos iteration
!     np -- (in) number of Lanczos vectors, the size of the main space
!     i  -- (out) current index of the Lanczos block vector
!     j  -- (out) current Lanczos vector in the block vector to be processed
c
c
      subroutine indexl( jj, np, j, i )
      implicit none

      integer, intent(in) :: jj, np
      integer, intent(out) :: i, j
      integer kk

C
C     AH1:LANCZ.INDEXL
C     THIS ROUTINE COMPUTES THE INDICES J,I OF THE JJ-TH LANCZOS-VECTOR.
C
      kk = jj / np                   ! Number of block Lanczos vector times matrix
                                     ! operations that have been processed so far
      i  = mod(kk+1, 3) + 1
      j  = jj + 1 - kk * np
      end subroutine indexl
      end subroutine hblanc
