!
! brillouinpathspec --- Read and write specifications of paths
!       through the Brillouin zone in the format of Wannier
! Christian Buth
! 10.07.2004
!
!
! Copyright (c) 2004-2006 Christian Buth
!
! This library is free software; you can redistribute it and/or
! modify it under the terms of the GNU Lesser General Public
! License as published by the Free Software Foundation; either
! version 2.1 of the License, or (at your option) any later version.
!
! This library is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! Lesser General Public License for more details.
!
! You should have received a copy of the GNU Lesser General Public
! License along with this library; if not, write to the Free Software
! Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
!
!
! Version 1.00: 10.07.2004 (Extracted code from wannier module)
! Version 1.10: 11.07.2004 (Added reporting path specifications)
! Version 1.20: 25.08.2004 (Length of Brillouin paths computed with the
!       metric of the reciprocal lattice)
! Version 1.30: 06.09.2004 (Direct initialization of Brillouin paths)
! Version 1.40: 22.02.2005 (Corrected bug in calculation of the
!       number of k-points, error checking for wrong increments
!       in Brillouin path specifications)
!


!
! Module
!

      module brillouinpathspec
      use fortrantl
      use lattice
      implicit none


!
! Types
!

      type tbrillouinpathspec
        integer numkpoints
        real(dblprec) metriclength, &           ! Length using the proper metric
     &                coordlength               ! Length in coordinate space
        type(tlatticevector) :: beg, end, inc
      end type tbrillouinpathspec


!
! Module procedures
!

      contains


!
! initbrillouinpathspec --- init Brillouin zone path information
!

      subroutine initbrillouinpathspec(bpath, periodicity, spacegroup, lbeg, lend, linc)
      implicit none

      integer, intent(in) :: periodicity, spacegroup
      type(tlatticevector), intent(in), optional :: lbeg, lend, linc
      type(tbrillouinpathspec), intent(inout) :: bpath
      integer pna, pnb, pnc
      real(dblprec) pnr, distvec(3)
      type(tlatticetraf) :: traf

      ! Get path information
      if (present(lbeg)) then
        bpath%beg = lbeg
      end if
      if (present(lend)) then
        bpath%end = lend
      end if
      if (present(lbeg)) then
        bpath%inc = linc
      end if

      ! Determine the number of kpoints along the path through the Brillouin zone
      pna = huge(pna)
      pnb = huge(pnb)
      pnc = huge(pnc)

!      if (bpath%inc%a /= 0) then
!        pna = (bpath%end%a - bpath%beg%a) / abs(bpath%inc%a)
!        bpath%inc%a = sign(bpath%inc%a, real(pna, dblprec))
!        pna = abs(pna)
!      end if
      if (bpath%inc%v(1) /= 0) then
        pnr = (bpath%end%v(1) - bpath%beg%v(1)) / abs(bpath%inc%v(1))
        bpath%inc%v(1) = sign(bpath%inc%v(1), pnr)
        pna = nint(abs(pnr))
      end if
!      if (bpath%inc%b /= 0) then
!        pnb = (bpath%end%b - bpath%beg%b) / abs(bpath%inc%b)
!        bpath%inc%b = sign(bpath%inc%b, real(pnb, dblprec))
!        pnb = abs(pnb)
!      end if
      if (bpath%inc%v(2) /= 0) then
        pnr = (bpath%end%v(2) - bpath%beg%v(2)) / abs(bpath%inc%v(2))
        bpath%inc%v(2) = sign(bpath%inc%v(2), pnr)
        pnb = nint(abs(pnr))
      end if
!      if (bpath%inc%c /= 0) then
!        pnc = (bpath%end%c - bpath%beg%c) / abs(bpath%inc%c)
!        bpath%inc%c = sign(bpath%inc%c, real(pnc, dblprec))
!        pnc = abs(pnc)
!      end if
      if (bpath%inc%v(3) /= 0) then
        pnr = (bpath%end%v(3) - bpath%beg%v(3)) / abs(bpath%inc%v(3))
        bpath%inc%v(3) = sign(bpath%inc%v(3), pnr)
        pnc = nint(abs(pnr))
      end if

      ! Check whether the end point of the path is reached for all coordinates
      if ((pna /= pnb .and. .not. (pna /= huge(pna) .or. pnb /= huge(pnb))) .or. &
     &    (pna /= pnc .and. .not. (pna /= huge(pna) .or. pnc /= huge(pnc))) .or. &
     &    (pnb /= pnc .and. .not. (pnb /= huge(pnb) .or. pnc /= huge(pnc)))) then
        write (*,*) 'initbrillouinpathspec: increments for the directions a ', &
     &          pna, ', b ', pnb, ', c ', pnc, ' in the Brillouinpath do ',    &
     &          'not lead to the same number of k-points!'
        stop 'Error in initbrillouinpathspec!'
      end if
      bpath%numkpoints = min(pna, pnb, pnc) + 1
      write (*,*) 'Number of points along the path ', bpath%numkpoints

      ! Determine the length of the path in coordinate space
!      bpath%coordlength = sqrt((bpath%end%a - bpath%beg%a)**2 &
!     &                      + (bpath%end%b - bpath%beg%b)**2 &
!     &                      + (bpath%end%c - bpath%beg%c)**2)
      bpath%coordlength = sqrt((bpath%end%v(1) - bpath%beg%v(1))**2 &
     &                       + (bpath%end%v(2) - bpath%beg%v(2))**2 &
     &                       + (bpath%end%v(3) - bpath%beg%v(3))**2)

      ! Determine the length of the path in reciprocal space
!      distvec(1) = bpath%end%a - bpath%beg%a
!      distvec(2) = bpath%end%b - bpath%beg%b
!      distvec(3) = bpath%end%c - bpath%beg%c
      distvec = bpath%end%v - bpath%beg%v
      !distvec(1) = abs(bpath%end%a - bpath%beg%a)
      !distvec(2) = abs(bpath%end%b - bpath%beg%b)
      !distvec(3) = abs(bpath%end%c - bpath%beg%c)
      traf = getRecSpTrafPrimCell(periodicity, spacegroup)
     !!   traf%m =  traf%m * 2.0
      distvec = matmul(traf%m, distvec)
      bpath%metriclength = sqrt(dot_product(distvec, distvec))
      end subroutine initbrillouinpathspec


!
! exitbrillouinpathspec --- exit Brillouin zone path information
!

      subroutine exitbrillouinpathspec(bpath)
      implicit none

      type(tbrillouinpathspec), intent(inout) :: bpath
      end subroutine exitbrillouinpathspec


!
! readbrillouinpathspec --- read in Brillouin zone path information in the
!       format also used by the Wannier program
!

      subroutine readbrillouinpathspec(bpath, funit, periodicity, spacegroup)
      implicit none

      integer, intent(in) :: funit, periodicity, spacegroup
      type(tbrillouinpathspec), intent(inout) :: bpath

      ! Read specification of a path through the Brillouin zone
!      read (funit,*) bpath%beg%a, bpath%end%a, bpath%inc%a
      read (funit,*) bpath%beg%v(1), bpath%end%v(1), bpath%inc%v(1)
      if (periodicity > 1) then
!        read (funit,*) bpath%beg%b, bpath%end%b, bpath%inc%b
        read (funit,*) bpath%beg%v(2), bpath%end%v(2), bpath%inc%v(2)
      else
!        bpath%beg%b = 0
!        bpath%end%b = 0
!        bpath%inc%b = 0
!        bpath%beg%c = 0
!        bpath%end%c = 0
!        bpath%inc%c = 0
        bpath%beg%v(2) = 0 ; bpath%end%v(2) = 0 ; bpath%inc%v(2) = 0
        bpath%beg%v(3) = 0 ; bpath%end%v(3) = 0 ; bpath%inc%v(3) = 0
      end if
      if (periodicity > 2) then
!        read (funit,*) bpath%beg%c, bpath%end%c, bpath%inc%c
        read (funit,*) bpath%beg%v(3), bpath%end%v(3), bpath%inc%v(3)
      else
!        bpath%beg%c = 0
!        bpath%end%c = 0
!        bpath%inc%c = 0
        bpath%beg%v(3) = 0 ; bpath%end%v(3) = 0 ; bpath%inc%v(3) = 0
      end if

      call initbrillouinpathspec(bpath, periodicity, spacegroup)
      end subroutine readbrillouinpathspec


!
! writebrillouinpathspec --- write out brillouin zone path information in the
!       format also used by the Wannier program
!

      subroutine writebrillouinpathspec(bpath, funit, periodicity)
      implicit none

      integer, intent(in) :: funit, periodicity
      type(tbrillouinpathspec), intent(in) :: bpath

!      write (funit,*) bpath%beg%a, bpath%end%a, &
!     &                bpath%inc%a
      write (funit,*) bpath%beg%v(1), bpath%end%v(1), &
     &                bpath%inc%v(1)
      if (periodicity > 1) then
!        write (funit,*) bpath%beg%b, bpath%end%b, &
!     &                  bpath%inc%b
        write (funit,*) bpath%beg%v(2), bpath%end%v(2), &
     &                  bpath%inc%v(2)
      end if
      if (periodicity > 2) then
!        write (funit,*) bpath%beg%c, bpath%end%c, &
!     &                  bpath%inc%c
        write (funit,*) bpath%beg%v(3), bpath%end%v(3), &
     &                  bpath%inc%v(3)
      end if
      end subroutine writebrillouinpathspec
      end module brillouinpathspec
