!
! coadcfile --- Manage CO-ADC matrix in a file
! Christian Buth
! 15.06.2004
!
!
! Copyright (c) 2004-2006 Christian Buth
!
! This library is free software; you can redistribute it and/or
! modify it under the terms of the GNU Lesser General Public
! License as published by the Free Software Foundation; either
! version 2.1 of the License, or (at your option) any later version.
!
! This library is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! Lesser General Public License for more details.
!
! You should have received a copy of the GNU Lesser General Public
! License along with this library; if not, write to the Free Software
! Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
!
!
! Version 1.00: 15.06.2004 (Manipulate CO-ADC matrix files for molecules)
! Version 1.10: 27.06.2004 (Added direct lattice information)
!
!
! Communication file for the properties of the CO-ADC matrix
! Assumed structure of 'Filecomm':
!
!         signature               ! Identify communication file 'CO-ADC_V1.00'
!         main                    ! Dimension of main space of CO-ADC matrix, i.e.
!                                 ! the number of vectors to be used in a block
!                                 ! Lanczos operation
!         nocc                    ! Number of occupied bands
!         nrecd                   ! Number of records written to the file Filedia_1
!         nreco                   ! Number of records written to the file Fileoff_1
!         ndim                    ! Dimension of the ADC matrix
!         nbuf                    ! Buffer size in Fileoff_1
!         periodicity             ! 0: molecule, 1: polymer, 2: surface, 3: crystal
!         numdiag                 ! Number of diagonal matrix elements
!         numoff                  ! Number of off-diagonal matrix elements
!         threshold               ! Threshold for writing matrix elements
!         numlattice              ! Number of direct space lattice vectors
!         idirect(numlattice)     ! Direct space lattice indices
!         cdirect(numlattice)     ! Direct space lattice vectors
!
!
! with the types of the variables being
!
!         character(len=12) signature
!         integer main, nocc, nrecd, nreco, ndim, nbuf, periodicity, numdiag, numoff
!         real threshold
!         type(latticeindex) idirect
!         type(latticevector) cdirect
!
!
! File for the diagonal matrix elements of the CO-ADC matrix
! Assumed structure of 'Filedia_1':
!
!         diag(nbuf)              ! Diagonal matrix elements of the CO-ADC matrix
!                                 ! Extension to Francescos Tarantellis file format
!         idx(nbuf)               ! Chunk of column indices
!         latd(nbuf)              ! Direct lattice vectors
!         nwd                     ! Number of datums in the record
!
!
! with the types of the variables being
!
!         real(dblprec) diag
!         integer idx, latd, nwd
!
! for nrecd records; nrecd, nbuf are given in the communication file of the CO-ADC matrix.
!
!
! File for the off-diagonal matrix elements of the CO-ADC matrix
! Assumed structure of 'Fileoff_1':
!
!         off(nbuf)               ! Chunk of matrix elements of the CO-ADC matrix
!         iox(nbuf)               ! Chunk of column indices
!         jox(nbuf)               ! Chunk of row indices
!                                 ! Extension to Francescos Tarantellis file format
!         lato(nbuf)              ! Direct lattice vectors
!         nwo                     ! Number of datums in the record
!
! with the types of the variables being
!
!         real(dblprec) off
!         integer iox, jox, nwo
!
! for nreco records; nreco, nbuf are given in the communication file of the CO-ADC matrix.
!


!
! Module
!

      module coadcfile
      use fortrantl
      use lattice
      implicit none


!
! Parameters
!

      character(12), parameter :: coadcsignature = 'CO-ADC_V1.00'


!
! Types
!

      type tcoadcmatrix
        logical readfile                        ! Read file (.true.) or write file (.false.)

        integer fadccom,     &                  ! Unit number for the communication file for the CO-ADC matrix (Filecom)
     &          fadcdia,     &                  ! Unit number for the file of diagonal matrix elements (Filedia)
     &          fadcoff,     &                  ! Unit number for the file of off-diagonal matrix elements (Fileoff)
     &          periodicity, &                  ! Periodicity of the system, 0: molecule, 1: polymer, 2: surface, 3: crystal
     &          main,        &                  ! Dimension of the main space, the number of Lanczos vectors
     &          nocc,        &                  ! Number of occupied bands
     &          nrecd,       &                  ! Number of records in Filedia
     &          nreco,       &                  ! Number of records in Fileoff
     &          ndim,        &                  ! Dimension of the CO-ADC matrix
     &          nbuf,        &                  ! Buffer size of Fileoff
     &          nwd,         &                  ! Number of entries in the current record of Filedia
     &          nwo,         &                  ! Number of entries in the current record of Fileoff
     &          currrecd,    &                  ! Presently processed record of Filedia
     &          currreco,    &                  ! Presently processed record of Fileoff
     &          numdiag,     &                  ! Number of diagonal elements in Filedia while reading,
                                                ! Number of diagonal elements written to Filedia while writing
     &          numoff                          ! Number of off-diagonal elements in Fileoff while reading,
                                                ! Number of off-diagonal elements written to Fileoff while writing

        integer, pointer :: idx(:),  &
     &                      iox(:),  &
     &                      jox(:),  &
     &                      latd(:), &
     &                      lato(:)

        real threshold                          ! Threshold for writing matrix elements to disk

        real(dblprec), pointer :: off(:), diag(:)

        type(tlattice) lattice                  ! Direct lattice vectors if periodicity > 0
      end type tcoadcmatrix


      contains


!
! opencoadcfile --- open an initialize files for diagonal and off-diagonal
!               elements of the CO-ADC matrix
!

      subroutine opencoadcfile(am)
      implicit none

      type(tcoadcmatrix), intent(inout) :: am
      character(len(coadcsignature)) signature
      integer rld, rlo, i

      ! Write files
      if (.not. am%readfile) then

        ! Determine record lengths
        if (am%periodicity == 0) then
          rld = (8 + 4) * am%nbuf + 4
          rlo = (8 + 4 + 4) * am%nbuf + 4
        else
          rld = (8 + 4 + 4) * am%nbuf + 4
          rlo = (8 + 4 + 4 + 4) * am%nbuf + 4
        end if

        ! Initialize file writing
        allocate(am%diag(am%nbuf), am%idx(am%nbuf), &
     &           am%off(am%nbuf), am%iox(am%nbuf), am%jox(am%nbuf))
        if (am%periodicity > 0) then
          allocate(am%latd(am%nbuf), am%lato(am%nbuf))
          if (am%lattice%size > 0) then
            allocate(am%lattice%idirect(am%lattice%size), &
     &               am%lattice%cdirect(am%lattice%size), &
     &               am%lattice%factor (am%lattice%size))
          end if
        end if

        am%nrecd = 0
        am%nreco = 0
        am%nwd = 0
        am%nwo = 0
        am%currrecd = 0
        am%currreco = 0
        am%numdiag = 0
        am%numoff = 0

        write (*,*) 'IO buffer for CO-ADC matrix', am%nbuf

        ! Open files for the diagonal and the off-diagonal matrix
        ! elements of the ADC matrix
        open(unit=am%fadccom, file='Filecom', status='replace', &
     &       action='write', form='formatted', access='sequential')
        open(unit=am%fadcdia, file='Filedia_1', status='replace', &
     &       action='write', form='unformatted', access='direct', recl=rld)
        open(unit=am%fadcoff, file='Fileoff_1', status='replace', &
     &       action='write', form='unformatted', access='direct', recl=rlo)


      ! Read files
      else

        ! Read communication file for the CO-ADC matrix
        open(unit=am%fadccom, file='Filecom', status='old', &
     &       action='read', form='formatted', access='sequential')
        read (am%fadccom,*) signature
        if (signature /= coadcsignature) then
          write (*,*) 'Filecom is not a valid '//coadcsignature//' file!'
          stop 'coadcfile: Error in opencoadcfile'
        end if
        read (am%fadccom,*) am%main, am%nocc, am%nrecd, am%nreco, am%ndim, &
     &          am%nbuf, am%periodicity, am%numdiag, &
     &          am%numoff, am%threshold, am%lattice%size
        if (am%periodicity > 0 .and. am%lattice%size > 0) then
          allocate(am%lattice%idirect(am%lattice%size), &
     &             am%lattice%cdirect(am%lattice%size), &
     &             am%lattice.factor(am%lattice%size))
          read (am%fadccom,*) (am%lattice%idirect(i), i=1, am%lattice%size), &
     &                      (am%lattice%cdirect(i), i=1, am%lattice%size)
        end if
        close (am%fadccom)

        allocate(am%diag(am%nbuf), am%idx(am%nbuf), &
     &           am%off(am%nbuf), am%iox(am%nbuf), am%jox(am%nbuf))
        if (am%periodicity > 0) then
          allocate(am%latd(am%nbuf), am%lato(am%nbuf))
        end if

        am%nwd = 0
        am%nwo = 0
        am%currrecd = 0
        am%currreco = 0

        ! Determine record lengths
        if (am%periodicity == 0) then
          rld = (8 + 4) * am%nbuf + 4
          rlo = (8 + 4 + 4) * am%nbuf + 4
        else
          rld = (8 + 4 + 4) * am%nbuf + 4
          rlo = (8 + 4 + 4 + 4) * am%nbuf + 4
        end if

        ! Initialize file writing
        write (*,*) 'IO buffer for CO-ADC matrix', am%nbuf
        write (*,*) 'main, nocc, nrecd, nreco, ndim, recld, reclo, nbuf, periodicity, ', &
     &              'numdiag, numoff, threshold, lattice%size'
        write (*,*) am%main, am%nocc, am%nrecd, am%nreco, am%ndim, rld, rlo, am%nbuf, am%periodicity, &
     &              am%numdiag, am%numoff, am%threshold, am%lattice%size

        ! Open files for the diagonal and the off-diagonal matrix
        ! elements of the CO-ADC matrix
        open(unit=am%fadcdia, file='Filedia_1', status='old', &
     &       action='read',form='unformatted', access='direct', recl=rld)
        open(unit=am%fadcoff, file='Fileoff_1', status='old', &
     &       action='read', form='unformatted', access='direct', recl=rlo)

      end if
      end subroutine opencoadcfile


!
! closecoadcfile --- close files for diagonal and off-diagonal elements
!               of the CO-ADC matrix
!

      subroutine closecoadcfile(am)
      implicit none

      integer i
      type(tcoadcmatrix), intent(inout) :: am

      if (.not. am%readfile) then
        call writematrixbuffer(am, .true.)

        write (am%fadccom,*) coadcsignature
        write (am%fadccom,*) am%main, am%nocc, am%nrecd, am%nreco, am%ndim, &
     &                    am%nbuf, am%periodicity, am%numdiag, &
     &                    am%numoff, am%threshold, am%lattice%size
        if (am%periodicity > 0 .and. am%lattice%size > 0) then
          write (am%fadccom,*) (am%lattice%idirect(i), i=1, am%lattice%size), &
     &                       (am%lattice%cdirect(i), i=1, am%lattice%size)
        end if
        close (am%fadccom)

        write (*,*) 'IO buffer for CO-ADC matrix', am%nbuf
        write (*,*) 'main, nocc, nrecd, nreco, ndim, nbuf, periodicity, ', &
     &              'numdiag, numoff, threshold, lattice%size'
        write (*,*) am%main, am%nocc, am%nrecd, am%nreco, am%ndim, am%nbuf, am%periodicity, &
     &              am%numdiag, am%numoff, am%threshold, am%lattice%size
      end if
      close (am%fadcdia)
      close (am%fadcoff)

      ! Deallocate memory
      deallocate(am%diag, am%off, am%idx, am%iox, am%jox)
      if (am%periodicity > 0) then
        deallocate(am%latd, am%lato)
        if (am%lattice%size > 0) then
          deallocate(am%lattice%idirect, am%lattice%cdirect, am%lattice.factor)
        end if
      end if
      end subroutine closecoadcfile


!
! writeentry --- write matrix element to the CO-ADC matrix file
!

      subroutine writeentry(am, i, j, R, elem)
      implicit none

      integer, intent(in) :: i, j, R
      real(dblprec), intent(in) :: elem
      type(tcoadcmatrix), intent(inout) :: am

      if (i == j) then
        call writediagentry(am, i, R, elem)
      else
        call writeoffdiagentry(am, i, j, R, elem)
      end if
      end subroutine writeentry


!
! writediagentry --- write diagonal matrix element to the CO-ADC matrix file
!

      subroutine writediagentry(am, i, R, delem)
      implicit none

      integer, intent(in) :: i, R
      real(dblprec), intent(in) :: delem
      type(tcoadcmatrix), intent(inout) :: am

      if (abs(delem) < am%threshold) then
        return
      end if

      am%nwd = am%nwd + 1
      am%numdiag = am%numdiag + 1
      am%diag(am%nwd)  = delem
      am%idx(am%nwd) = i
      if (am%periodicity > 0) then
        am%latd(am%nwd)  = R
      end if

      if (am%nwd == am%nbuf) then
        call writematrixbuffer(am, .false.)
      end if
      end subroutine writediagentry


!
! writeoffdiagentry --- write off-diagonal matrix element to the CO-ADC matrix file
!

      subroutine writeoffdiagentry(am, i, j, R, oelem)
      implicit none

      integer, intent(in) :: i, j, R
      real(dblprec), intent(in) ::  oelem
      type(tcoadcmatrix), intent(inout) :: am

      if (abs(oelem) < am%threshold) then
        return
      end if

      am%nwo = am%nwo + 1
      am%numoff = am%numoff + 1
      am%off(am%nwo)  = oelem
      am%iox(am%nwo) = i
      am%jox(am%nwo) = j
      if (am%periodicity > 0) then
        am%lato(am%nwo)  = R
      end if

      if (am%nwo == am%nbuf) then
        call writematrixbuffer(am, .false.)
      end if
      end subroutine writeoffdiagentry


!
! writematrixbuffer --- write chunk of off-diagonal matrix elements to the
!               CO-ADC matrix file
!

      subroutine writematrixbuffer(am, forceflush)
      implicit none

      logical, intent(in) :: forceflush
      type(tcoadcmatrix), intent(inout) :: am

      ! Flush buffer of the diagonal matrix elements of the CO-ADC matrix
      if (forceflush .or. am%nwd == am%nbuf) then
        am%nrecd = am%nrecd + 1
        if (am%periodicity == 0) then
          write (unit=am%fadcdia, rec=am%nrecd) am%diag, am%idx, am%nwd
        else
          write (unit=am%fadcdia, rec=am%nrecd) am%diag, am%idx, am%latd, am%nwd
        end if
        am%nwd = 0
      end if

      ! Flush buffer of the off-diagonal matrix elements of the CO-ADC matrix
      if (forceflush .or. am%nwo == am%nbuf) then
        am%nreco = am%nreco + 1
        if (am%periodicity == 0) then
          write (unit=am%fadcoff, rec=am%nreco) am%off, am%iox, am%jox, am%nwo
        else
          write (unit=am%fadcoff, rec=am%nreco) am%off, am%iox, am%jox, am%lato, am%nwo
        end if
        am%nwo = 0
      end if
      end subroutine writematrixbuffer


!
! readmatrixbuffer --- read chunk of off-diagonal matrix elements to the
!               adc matrix file
!

      subroutine readmatrixbuffer(am, diagoff, ameof)
      implicit none

      logical, intent(in) :: diagoff
      logical, intent(out) :: ameof
      type(tcoadcmatrix), intent(inout) :: am

      if (diagoff) then

        am%currrecd = am%currrecd + 1
        if (am%currrecd > am%nrecd) then
          ameof = .true.
        else
          if (am%periodicity == 0) then
            read (unit=am%fadcdia, rec=am%currrecd) am%diag, am%idx, am%nwd
          else
            read (unit=am%fadcdia, rec=am%currrecd) am%diag, am%idx, am%latd, am%nwd
          end if
          ameof = .false.
        end if

      else

        am%currreco = am%currreco + 1
        if (am%currreco > am%nreco) then
          ameof = .true.
        else
          if (am%periodicity == 0) then
            read (unit=am%fadcoff, rec=am%currreco) am%off, am%iox, am%jox, am%nwo
          else
            read (unit=am%fadcoff, rec=am%currreco) am%off, am%iox, am%jox, am%lato, am%nwo
          end if
          ameof = .false.
        end if

      end if
      end subroutine readmatrixbuffer


!
! amrewind --- rewind files for rereading
!

      subroutine amrewind(am)
      implicit none

      type(tcoadcmatrix), intent(inout) :: am

      am%currrecd = 0
      am%currreco = 0
      am%nwd = 0
      am%nwo = 0

      rewind(am%fadcdia)
      rewind(am%fadcoff)
      end subroutine amrewind
      end module coadcfile
