!
! lattice --- Provide types for crystal lattices
! Christian Buth
! 27.02.2004
!
!
! Copyright (c) 2004-2006 Christian Buth
!
! This library is free software; you can redistribute it and/or
! modify it under the terms of the GNU Lesser General Public
! License as published by the Free Software Foundation; either
! version 2.1 of the License, or (at your option) any later version.
!
! This library is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
! Lesser General Public License for more details.
!
! You should have received a copy of the GNU Lesser General Public
! License along with this library; if not, write to the Free Software
! Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
!
!
! Version 1.00: 27.02.2004 (Lattice index and lattice vectors)
! Version 1.10: 10.07.2004 (Direct lattices, lattice factors)
! Version 1.20: 25.08.2004 (Transformation to primitive unit cells)
! Version 1.30: 22.02.2005 (Changed definition of tlatticeindex and
!       tlatticevector from a, b, c to i(3) and v(3), respectively)
!


!
! Module
!

      module lattice
      use fortrantl
      implicit none


!
! Types
!

      type tlatticeindex
!        integer a, b, c
        integer i(3)
      end type tlatticeindex


      type tlatticevector
!        real(dblprec) a, b, c
        real(dblprec) v(3)
      end type tlatticevector


      type tlatticetraf
        real(dblprec) m(3,3)
      end type tlatticetraf


      type tlattice
        integer size, spacegroup
        complex(dblprec), pointer :: factor(:)
        type(tlatticevector) crystalmomentum
        type(tlatticeindex), pointer :: idirect(:)
        type(tlatticevector), pointer :: cdirect(:)
      end type tlattice


!
! Procedures
!

      contains


!
! latticefactor --- compute the e^{ikR} factor for a certain k point and a certain
!       direct space lattice vector R
!

      complex(dblprec) function latticefactor(k, R)
      use mathtl
      implicit none

      real(dblprec) kR
      type(tlatticevector), intent(in) :: k, R

!      kR = 2.0 * M_PI * (k%a * R%a + k%b * R%b + k%c * R%c)
      kR = 2.0 * M_PI * dot_product(k%v, R%v)
      latticefactor = cmplx(cos(kR), sin(kR), dblprec)
      end function latticefactor


!
! newk --- compute e^{ikR} factors for a certain k point
!

      subroutine newk(lat, k)
      implicit none

      type(tlatticevector), intent(in) :: k
      type(tlattice), intent(inout) :: lat
      integer R

      lat%crystalmomentum = k
      do R=1, lat%size
        lat%factor(R) = latticefactor(k, lat%cdirect(R))
      end do
      end subroutine newk


!
! getRecSpTrafConvCell --- get the transformation matrix from the
!       conventional cell in the reciprocal lattice to the primitive
!       cell in the reciprocal lattice
!

      type(tlatticetraf) function getRecSpTrafConvCell(periodicity, spacegroup)
      use fortrantl
      implicit none

      integer, intent(in) :: periodicity, spacegroup

      select case(periodicity)
        case (1)
          getRecSpTrafConvCell%m = 0
          getRecSpTrafConvCell%m(1,1) = 1
          getRecSpTrafConvCell%m(2,2) = 1
          getRecSpTrafConvCell%m(3,3) = 1
        case (3)
          getRecSpTrafConvCell%m      = 0.5
          getRecSpTrafConvCell%m(1,1) = 0
          getRecSpTrafConvCell%m(2,2) = 0
          getRecSpTrafConvCell%m(3,3) = 0
!          stop 'More code here (lattice.F90)'
        case default
          stop 'periodicity not supported'
      end select
      end function getRecSpTrafConvCell


!
! getRecSpTrafPrimCell --- get the transformation matrix from the
!       primitive cell in the reciprocal lattice to the conventional
!       cell in the reciprocal lattice
!

      type(tlatticetraf) function getRecSpTrafPrimCell(periodicity, spacegroup)
      use fortrantl
      implicit none

      integer, intent(in) :: periodicity, spacegroup

      select case(periodicity)
        case (1)
          getRecSpTrafPrimCell%m = 0
          getRecSpTrafPrimCell%m(1,1) = 1
          getRecSpTrafPrimCell%m(2,2) = 1
          getRecSpTrafPrimCell%m(3,3) = 1
        case (3)
          getRecSpTrafPrimCell%m = 1
          getRecSpTrafPrimCell%m(1,1) = -1
          getRecSpTrafPrimCell%m(2,2) = -1
          getRecSpTrafPrimCell%m(3,3) = -1
!          stop 'More code here (lattice.F90)'
        case default
          stop 'periodicity not supported'
      end select
      end function getRecSpTrafPrimCell
      end module lattice
